#include <inc/fs.h>
#include <inc/string.h>
#include <inc/lib.h>

#define debug 0

struct FileDev : public Dev {
	FileDev() : Dev('f', "file") {
	}
	
	ssize_t read(Fd *fd, void *buf, size_t n, off_t offset);
	ssize_t write(Fd *fd, const void *buf, size_t n, off_t offset);
	int close(Fd *fd);
	int stat(Fd *fd, struct Stat *stat);
	int trunc(Fd *fd, off_t newsize);
};

FileDev file_dev;

extern "C" {
extern Fsreq fsreqbuf;		// page-aligned, declared in entry.S
}

// Send an IP request to the file server, and wait for a reply.
// type: request code, passed as the simple integer IPC value.
// fileid: file ID (or -1 if no fileid necessary).
// offset: offset/open mode (or 0 if no offset or open mode necessary).
// path: pathname (or NULL if no pathname necessary).
// dstva: virtual address at which to receive reply page, 0 if none.
// *perm: permissions of received page.
// Returns 0 if successful, < 0 on failure.
// Error messages are the same as for ipc_recv, plus:
//   -E_BAD_PATH  if strlen(path) >= MAXPATHLEN.
int
fsipc(unsigned type, int fileid, off_t offset, const char *path,
      void *dstva, int *perm)
{
	fsreqbuf.req_fileid = fileid;
	fsreqbuf.req_offset = offset;
	if (path) {
		if (strlen(path) >= MAXPATHLEN)
			return -E_BAD_PATH;
		strcpy(fsreqbuf.req_path, path);
	}
	
	if (debug)
		cprintf("[%08x] fsipc %d\n", env->env_id, type);

	ipc_send(envs[0].env_id, type, &fsreqbuf, PTE_P | PTE_W | PTE_U);
	envid_t whom;
	return ipc_recv(&whom, dstva, perm);
}

// Helper functions for file access
static int fmap(Fd *fd, off_t oldsize, off_t newsize);
static int funmap(Fd *fd, off_t oldsize, off_t newsize, bool dirty);

// Open a file (or directory),
// returning the file descriptor index on success, < 0 on failure.
int
open(const char *path, int omode)
{
	// Find an unused file descriptor page using Fd::alloc.
	// Then send a message to the file server to open a file using the
	// fsipc() function.
	// The client-to-server FSREQ_OPEN message has the following format:
	// - "fileid" is -1.
	// - "offset" is the open mode.
	// - "path" is the pathname to open.
	// In reply, the server creates a memory page containing the opened
	// file descriptor and uses IPC to map that page into the client's
	// address space.
	// Note that Fd::alloc does not allocate a page, it just returns an
	// unused fd address.  You'll use that address as an argument to
	// fsipc()!
	// Finally, map the file data using fmap() and return the file
	// descriptor number (Fd::fd_number()).
	// If any step fails, use close to free the file descriptor.

	// LAB 5: Your code here.
	panic("open() unimplemented!");
}

// Clean up a file-server file descriptor.
// This function is called by close.
int
FileDev::close(Fd *fdx)
{
	// Unmap any data mapped for the file,
	// then use fsipc to tell the file server that we have closed the file
	// (to free up its resources).

	// LAB 5: Your code here.
	panic("close() unimplemented!");
}

// Read 'n' bytes from 'fd' at the current seek position into 'buf'.
// Since files are memory-mapped, this amounts to a memcpy()
// surrounded by a little red tape to handle the file size and seek pointer.
ssize_t
FileDev::read(Fd *fdx, void *buf, size_t n, off_t offset)
{
	FileFd *fd = (FileFd *) fdx;
	off_t size;

	// avoid reading past the end of file
	size = fd->fd_file.f_size;
	if (offset > size)
		return 0;
	if ((off_t) (offset + n) > size)
		n = size - offset;

	// read the data by copying from the file mapping
	memcpy(buf, fd->data() + offset, n);
	return n;
}

// Find the page that maps the file block starting at 'offset',
// and store its address in '*blk'.
int
read_map(int fdnum, off_t offset, void **blk)
{
	int r;
	char *va;
	Fd *fd;

	if (!(fd = Fd::fd(fdnum, true)))
		return -E_INVAL;
	if (fd->fd_type != file_dev.dev_id)
		return -E_INVAL;
	va = fd->data() + offset;
	if (offset >= MAXFILESIZE)
		return -E_NO_DISK;
	if (!(vpd[PDX(va)] & PTE_P) || !(vpt[PGNUM(va)] & PTE_P))
		return -E_NO_DISK;
	*blk = (void*) va;
	return 0;
}

// Write 'n' bytes from 'buf' to 'fd' at the current seek position.
ssize_t
FileDev::write(Fd *fdx, const void *buf, size_t n, off_t offset)
{
	FileFd *fd = (FileFd *) fdx;
	int r;
	off_t tot;

	// don't write past the maximum file size
	tot = offset + n;
	if (tot > MAXFILESIZE)
		return -E_NO_DISK;

	// increase the file's size if necessary
	if (tot > fd->fd_file.f_size) {
		if ((r = trunc(fd, tot)) < 0)
			return r;
	}

	// write the data
	memcpy(fd->data() + offset, buf, n);
	return n;
}

int
FileDev::stat(Fd *fdx, struct Stat *st)
{
	FileFd *fd = (FileFd *) fdx;
	strcpy(st->st_name, fd->fd_file.f_name);
	st->st_size = fd->fd_file.f_size;
	st->st_isdir = (fd->fd_file.f_type == FTYPE_DIR);
	return 0;
}

// Truncate or extend an open file to 'size' bytes
int
FileDev::trunc(Fd *fdx, off_t newsize)
{
	FileFd *fd = (FileFd *) fdx;
	int r;

	if (newsize > MAXFILESIZE)
		return -E_NO_DISK;

	uint32_t fileid = fd->fd_fileid;
	off_t oldsize = fd->fd_file.f_size;
	if ((r = fsipc(FSREQ_SET_SIZE, fileid, newsize, NULL, NULL, NULL)) < 0)
		return r;
	assert(fd->fd_file.f_size == newsize);

	if ((r = fmap(fd, oldsize, newsize)) < 0)
		return r;
	funmap(fd, oldsize, newsize, 0);

	return 0;
}

// Call the file system server to obtain and map file pages
// when the size of the file as mapped in our memory increases.
// Harmlessly does nothing if oldsize >= newsize.
// Returns 0 on success, < 0 on error.
// If there is an error, unmaps any newly allocated pages.
static int
fmap(Fd *fdx, off_t oldsize, off_t newsize)
{
	FileFd *fd = (FileFd *) fdx;
	int r, perm;
	
	char *va = fd->data();
	for (off_t i = round_up(oldsize, PGSIZE); i < newsize; i += PGSIZE) {
		if ((r = fsipc(FSREQ_MAP, fd->fd_fileid, i, NULL,
			       va + i, &perm)) < 0) {
			// unmap anything we may have mapped so far
			funmap(fd, i, oldsize, 0);
			return r;
		} else if ((perm & ~(PTE_W | PTE_SHARE)) != (PTE_P | PTE_U))
			panic("fmap: unexpected permissions %08x for dstva %08x", perm, va + i);
	}
	return 0;
}

// Unmap any file pages that no longer represent valid file pages
// when the size of the file as mapped in our address space decreases.
// Harmlessly does nothing if newsize >= oldsize.
static int
funmap(Fd *fdx, off_t oldsize, off_t newsize, bool dirty)
{
	FileFd *fd = (FileFd *) fdx;
	int r;

	char *va = fd->data();

	// Check vpd to see if anything is mapped.
	if (!(vpd[PDX(va)] & PTE_P))
		return 0;

	int ret = 0;
	for (off_t i = round_up(newsize, PGSIZE); i < oldsize; i += PGSIZE)
		if (vpt[PGNUM(va + i)] & PTE_P) {
			if (dirty
			    && (vpt[PGNUM(va + i)] & PTE_D)
			    && (r = fsipc(FSREQ_DIRTY, fd->fd_fileid, i, NULL,
					  NULL, NULL)) < 0)
				ret = r;
			sys_page_unmap(0, va + i);
		}
  	return ret;
}

// Delete a file
int
remove(const char *path)
{
	return fsipc(FSREQ_REMOVE, -1, 0, path, NULL, NULL);
}

// Synchronize disk with buffer cache
int
sync(void)
{
	return fsipc(FSREQ_SYNC, -1, 0, NULL, NULL, NULL);
}

